@@ -17,8 +17,11 @@ const Mainloop = imports.mainloop;
 const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const St = imports.gi.St;
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
 const Chrome = imports.ui.chrome;
+const CtrlAltTab = imports.ui.ctrlAltTab;
 const Environment = imports.ui.environment;
 const ExtensionSystem = imports.ui.extensionSystem;
 const MessageTray = imports.ui.messageTray;
@@ -50,6 +53,7 @@ let messageTray = null;
 let notificationDaemon = null;
 let windowAttentionHandler = null;
 let telepathyClient = null;
+let ctrlAltTabManager = null;
 let recorder = null;
 let shellDBusService = null;
 let modalCount = 0;
@@ -135,6 +139,10 @@ function start() {
     notificationDaemon = new NotificationDaemon.NotificationDaemon();
     windowAttentionHandler = new WindowAttentionHandler.WindowAttentionHandler();
     telepathyClient = new TelepathyClient.Client();
+    panel.startStatusArea();
+
+    ctrlAltTabManager = new CtrlAltTab.CtrlAltTabManager();
+    ctrlAltTabManager.addGroup(panel.actor, _("Panel"), 'gnome-panel');
 
     _startDate = new Date();
 
@@ -329,6 +337,14 @@ function _globalKeyPressHandler(actor, event) {
         case Meta.KeyBindingAction.COMMAND_2:
             getRunDialog().open();
             return true;
+        case Meta.KeyBindingAction.SWITCH_PANELS:
+            // Only intercept this when we're in the overview, and don't
+            // intercept it if something beyond that (like, say, the
+            // ctrl-alt-tab popup!) is visible
+            if (overview.visible && modalCount == 1) {
+                ctrlAltTabManager.popup(modifierState & Clutter.ModifierType.SHIFT_MASK);
+                return true;
+            }
     }
 
     return false;
@@ -349,13 +365,13 @@ function _findModal(actor) {
  * @actor: #ClutterActor which will be given keyboard focus
  *
  * Ensure we are in a mode where all keyboard and mouse input goes to
- * the stage.  Multiple calls to this function act in a stacking fashion;
- * the effect will be undone when an equal number of popModal() invocations
- * have been made.
+ * the stage, and focus @actor. Multiple calls to this function act in
+ * a stacking fashion; the effect will be undone when an equal number
+ * of popModal() invocations have been made.
  *
- * Next, record the current Clutter keyboard focus on a stack.  If the modal stack
- * returns to this actor, reset the focus to the actor which was focused
- * at the time pushModal() was invoked.
+ * Next, record the current Clutter keyboard focus on a stack. If the
+ * modal stack returns to this actor, reset the focus to the actor
+ * which was focused at the time pushModal() was invoked.
  *
  * Returns: true iff we successfully acquired a grab or already had one
  */
@@ -385,7 +401,7 @@ function pushModal(actor) {
     }
     modalActorFocusStack.push([actor, curFocus]);
 
-    global.stage.set_key_focus(null);
+    global.stage.set_key_focus(actor);
     return true;
 }
 
@@ -415,6 +431,7 @@ function popModal(actor) {
     if (modalCount > 0)
         return;
 
+    global.stage.set_key_focus(null);
     global.end_modal(global.get_current_time());
     global.set_stage_input_mode(Shell.StageInputMode.NORMAL);
 }
