@@ -87,7 +87,7 @@ PopupBaseMenuItem.prototype = {
     },
 
     _onButtonReleaseEvent: function (actor, event) {
-        this.emit('activate', event);
+        this.activate(event);
         return true;
     },
 
@@ -95,7 +95,7 @@ PopupBaseMenuItem.prototype = {
         let symbol = event.get_key_symbol();
 
         if (symbol == Clutter.KEY_space || symbol == Clutter.KEY_Return) {
-            this.emit('activate', event);
+            this.activate(event);
             return true;
         }
         return false;
@@ -132,29 +132,16 @@ PopupBaseMenuItem.prototype = {
         this.emit('destroy');
     },
 
-    // true if non descendant content includes @actor
-    contains: function(actor) {
-        return false;
-    },
-
-    // adds an actor to the menu item; @column defaults to the next
-    // open column, @span defaults to 1. If @span is -1, the actor
-    // will span the width of the menu item. Children are not
-    // allowed to overlap each other.
-    addActor: function(child, column, span) {
-        if (column == null) {
-            if (this._children.length) {
-                let lastChild = this._children[this._children.length - 1];
-                column = lastChild.column + lastChild.span;
-            } else
-                column = 0;
-            span = 1;
-        } else if (span == null)
-            span = 1;
-
-        this._children.push({ actor: child,
-                              column: column,
-                              span: span });
+    // adds an actor to the menu item; @params can contain %span
+    // (column span; defaults to 1, -1 means "all the remaining width"),
+    // %expand (defaults to #false), and %align (defaults to
+    // #St.Align.START)
+    addActor: function(child, params) {
+        params = Params.parse(params, { span: 1,
+                                        expand: false,
+                                        align: St.Align.START });
+        params.actor = child;
+        this._children.push(params);
         this.actor.connect('destroy', Lang.bind(this, function () { this._removeChild(child); }));
         this.actor.add_actor(child);
     },
@@ -272,25 +259,44 @@ PopupBaseMenuItem.prototype = {
         for (let i = 0, col = 0; i < this._children.length; i++) {
             let child = this._children[i];
             let childBox = new Clutter.ActorBox();
-            childBox.x1 = x;
+
+            let [minWidth, naturalWidth] = child.actor.get_preferred_width(-1);
+            let availWidth, extraWidth;
             if (this._columnWidths) {
                 if (child.span == -1)
-                    childBox.x2 = box.x2;
+                    availWidth = box.x2 - x;
                 else {
-                    childBox.x2 = x;
+                    availWidth = 0;
                     for (let j = 0; j < child.span; j++)
-                        childBox.x2 += this._columnWidths[col++];
+                        availWidth += this._columnWidths[col++];
                 }
+                extraWidth = availWidth - naturalWidth;
             } else {
-                let [min, natural] = child.actor.get_preferred_width(-1);
-                childBox.x2 = x + natural;
+                availWidth = naturalWidth;
+                extraWidth = 0;
             }
-            let [min, natural] = child.actor.get_preferred_height(-1);
-            childBox.y1 = Math.round(box.y1 + (height - natural) / 2);
-            childBox.y2 = childBox.y1 + natural;
+
+            if (child.expand) {
+                childBox.x1 = x;
+                childBox.x2 = x + availWidth;
+            } else if (child.align === St.Align.CENTER) {
+                childBox.x1 = x + Math.round(extraWidth / 2);
+                childBox.x2 = childBox.x1 + naturalWidth;
+            } else if (child.align === St.Align.END) {
+                childBox.x2 = x + availWidth;
+                childBox.x1 = childBox.x2 - naturalWidth;
+            } else {
+                childBox.x1 = x;
+                childBox.x2 = x + naturalWidth;
+            }
+
+            let [minHeight, naturalHeight] = child.actor.get_preferred_height(-1);
+            childBox.y1 = Math.round(box.y1 + (height - naturalHeight) / 2);
+            childBox.y2 = childBox.y1 + naturalHeight;
+
             child.actor.allocate(childBox, flags);
 
-            x = childBox.x2 + this._spacing;
+            x += availWidth + this._spacing;
         }
     }
 };
@@ -322,7 +328,7 @@ PopupSeparatorMenuItem.prototype = {
         PopupBaseMenuItem.prototype._init.call(this, { reactive: false });
 
         this._drawingArea = new St.DrawingArea({ style_class: 'popup-separator-menu-item' });
-        this.addActor(this._drawingArea, 0, -1);
+        this.addActor(this._drawingArea, { span: -1, expand: true });
         this._drawingArea.connect('repaint', Lang.bind(this, this._onRepaint));
     },
 
@@ -367,7 +373,7 @@ PopupSliderMenuItem.prototype = {
         this._value = Math.max(Math.min(value, 1), 0);
 
         this._slider = new St.DrawingArea({ style_class: 'popup-slider-menu-item', reactive: true });
-        this.addActor(this._slider, 0, -1);
+        this.addActor(this._slider, { span: -1, expand: true });
         this._slider.connect('repaint', Lang.bind(this, this._sliderRepaint));
         this._slider.connect('button-press-event', Lang.bind(this, this._startDragging));
         this.actor.connect('scroll-event', Lang.bind(this, this._onScrollEvent));
@@ -534,7 +540,7 @@ PopupSwitchMenuItem.prototype = {
         this._switch = new Switch(active);
 
         this.addActor(this.label);
-        this.addActor(this._switch.actor);
+        this.addActor(this._switch.actor, { align: St.Align.END });
 
         this.connect('activate', Lang.bind(this,function(from) {
             this.toggle();
@@ -569,7 +575,7 @@ PopupImageMenuItem.prototype = {
         this.label = new St.Label({ text: text });
         this.addActor(this.label);
         this._icon = new St.Icon({ style_class: 'popup-menu-icon' });
-        this.addActor(this._icon);
+        this.addActor(this._icon, { align: St.Align.END });
 
         this.setIcon(iconName);
     },
@@ -601,76 +607,21 @@ function findNextInCycle(items, current, direction) {
     return items[mod(cur + direction, items.length)];
 }
 
-function PopupMenu() {
-    this._init.apply(this, arguments);
+function PopupMenuBase() {
+    throw new TypeError('Trying to instantiate abstract class PopupMenuBase');
 }
 
-PopupMenu.prototype = {
-    _init: function(sourceActor, alignment, arrowSide, gap) {
+PopupMenuBase.prototype = {
+    _init: function(sourceActor, styleClass) {
         this.sourceActor = sourceActor;
-        this._alignment = alignment;
-        this._arrowSide = arrowSide;
-        this._gap = gap;
 
-        this._boxPointer = new BoxPointer.BoxPointer(arrowSide,
-                                                     { x_fill: true,
-                                                       y_fill: true,
-                                                       x_align: St.Align.START });
-        this.actor = this._boxPointer.actor;
-        this.actor.style_class = 'popup-menu-boxpointer';
-        this._boxWrapper = new Shell.GenericContainer();
-        this._boxWrapper.connect('get-preferred-width', Lang.bind(this, this._boxGetPreferredWidth));
-        this._boxWrapper.connect('get-preferred-height', Lang.bind(this, this._boxGetPreferredHeight));
-        this._boxWrapper.connect('allocate', Lang.bind(this, this._boxAllocate));
-        this._boxPointer.bin.set_child(this._boxWrapper);
-
-        this._box = new St.BoxLayout({ style_class: 'popup-menu-content',
-                                       vertical: true });
-        this._boxWrapper.add_actor(this._box);
-        this.actor.add_style_class_name('popup-menu');
-
-        global.focus_manager.add_group(this.actor);
-
-        if (sourceActor._delegate instanceof PopupSubMenuMenuItem) {
-            this._isSubMenu = true;
-            this.actor.reactive = true;
-            this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
-        }
+        this.box = new St.BoxLayout({ style_class: styleClass,
+                                      vertical: true });
 
         this.isOpen = false;
         this._activeMenuItem = null;
     },
 
-    _boxGetPreferredWidth: function (actor, forHeight, alloc) {
-        // Update the menuitem column widths
-        let columnWidths = [];
-        let items = this._box.get_children();
-        for (let i = 0; i < items.length; i++) {
-            if (items[i]._delegate instanceof PopupBaseMenuItem) {
-                let itemColumnWidths = items[i]._delegate.getColumnWidths();
-                for (let j = 0; j < itemColumnWidths.length; j++) {
-                    if (j >= columnWidths.length || itemColumnWidths[j] > columnWidths[j])
-                        columnWidths[j] = itemColumnWidths[j];
-                }
-            }
-        }
-        for (let i = 0; i < items.length; i++) {
-            if (items[i]._delegate instanceof PopupBaseMenuItem)
-                items[i]._delegate.setColumnWidths(columnWidths);
-        }
-
-        // Now they will request the right sizes
-        [alloc.min_size, alloc.natural_size] = this._box.get_preferred_width(forHeight);
-    },
-
-    _boxGetPreferredHeight: function (actor, forWidth, alloc) {
-        [alloc.min_size, alloc.natural_size] = this._box.get_preferred_height(forWidth);
-    },
-
-    _boxAllocate: function (actor, box, flags) {
-        this._box.allocate(box, flags);
-    },
-
     addAction: function(title, callback) {
         var menuItem = new PopupMenuItem(title);
         this.addMenuItem(menuItem);
@@ -681,9 +632,29 @@ PopupMenu.prototype = {
 
     addMenuItem: function(menuItem, position) {
         if (position == undefined)
-            this._box.add(menuItem.actor);
+            this.box.add(menuItem.actor);
         else
-            this._box.insert_actor(menuItem.actor, position);
+            this.box.insert_actor(menuItem.actor, position);
+        if (menuItem instanceof PopupSubMenuMenuItem) {
+            if (position == undefined)
+                this.box.add(menuItem.menu.actor);
+            else
+                this.box.insert_actor(menuItem.menu.actor, position + 1);
+            menuItem._subMenuActivateId = menuItem.menu.connect('activate', Lang.bind(this, function() {
+                this.emit('activate');
+                this.close(true);
+            }));
+            menuItem._subMenuActiveChangeId = menuItem.menu.connect('active-changed', Lang.bind(this, function(submenu, submenuItem) {
+                if (this._activeMenuItem && this._activeMenuItem != submenuItem)
+                    this._activeMenuItem.setActive(false);
+                this._activeMenuItem = submenuItem;
+                this.emit('active-changed', submenuItem);
+            }));
+            menuItem._closingId = this.connect('open-state-changed', function(self, open) {
+                if (!open)
+                    menuItem.menu.close(false);
+            });
+        }
         menuItem._activeChangeId = menuItem.connect('active-changed', Lang.bind(this, function (menuItem, active) {
             if (active && this._activeMenuItem != menuItem) {
                 if (this._activeMenuItem)
@@ -697,22 +668,50 @@ PopupMenu.prototype = {
         }));
         menuItem._activateId = menuItem.connect('activate', Lang.bind(this, function (menuItem, event) {
             this.emit('activate', menuItem);
-            this.close();
+            this.close(true);
         }));
         menuItem.connect('destroy', Lang.bind(this, function(emitter) {
             menuItem.disconnect(menuItem._activateId);
             menuItem.disconnect(menuItem._activeChangeId);
+            if (menuItem.menu) {
+                menuItem.menu.disconnect(menuItem._subMenuActivateId);
+                menuItem.menu.disconnect(menuItem._subMenuActiveChangeId);
+                this.disconnect(menuItem._closingId);
+            }
             if (menuItem == this._activeMenuItem)
                 this._activeMenuItem = null;
         }));
     },
 
+    getColumnWidths: function() {
+        let columnWidths = [];
+        let items = this.box.get_children();
+        for (let i = 0; i < items.length; i++) {
+            if (items[i]._delegate instanceof PopupBaseMenuItem || items[i]._delegate instanceof PopupMenuBase) {
+                let itemColumnWidths = items[i]._delegate.getColumnWidths();
+                for (let j = 0; j < itemColumnWidths.length; j++) {
+                    if (j >= columnWidths.length || itemColumnWidths[j] > columnWidths[j])
+                        columnWidths[j] = itemColumnWidths[j];
+                }
+            }
+        }
+        return columnWidths;
+    },
+
+    setColumnWidths: function(widths) {
+        let items = this.box.get_children();
+        for (let i = 0; i < items.length; i++) {
+            if (items[i]._delegate instanceof PopupBaseMenuItem || items[i]._delegate instanceof PopupMenuBase)
+                items[i]._delegate.setColumnWidths(widths);
+        }
+    },
+
     addActor: function(actor) {
-        this._box.add(actor);
+        this.box.add(actor);
     },
 
     getMenuItems: function() {
-        return this._box.get_children().map(function (actor) { return actor._delegate; });
+        return this.box.get_children().map(function (actor) { return actor._delegate; }).filter(function(item) { return item instanceof PopupBaseMenuItem; });
     },
 
     removeAll: function() {
@@ -723,110 +722,22 @@ PopupMenu.prototype = {
         }
     },
 
-    setArrowOrigin: function(origin) {
-        this._boxPointer.setArrowOrigin(origin);
-    },
-
     activateFirst: function() {
-        let children = this._box.get_children();
+        let children = this.box.get_children();
         for (let i = 0; i < children.length; i++) {
             let actor = children[i];
-            if (actor._delegate && actor.visible && actor.reactive) {
+            if (actor._delegate && actor._delegate instanceof PopupBaseMenuItem && actor.visible && actor.reactive) {
                 actor._delegate.setActive(true);
                 break;
             }
         }
     },
 
-    open: function() {
-        if (this.isOpen)
-            return;
-
-        let primary = global.get_primary_monitor();
-
-        // We need to show it now to force an allocation,
-        // so that we can query the correct size.
-        this.actor.show();
-
-        // Position correctly relative to the sourceActor
-        let [sourceX, sourceY] = this.sourceActor.get_transformed_position();
-        let [sourceWidth, sourceHeight] = this.sourceActor.get_transformed_size();
-
-        let [minWidth, minHeight, natWidth, natHeight] = this.actor.get_preferred_size();
-
-        let menuWidth = natWidth, menuHeight = natHeight;
-
-        // Position the non-pointing axis
-        if (this._isSubmenu) {
-            if (this._arrowSide == St.Side.TOP || this._arrowSide == St.Side.BOTTOM) {
-                // vertical submenu
-                if (sourceY + sourceHeigth + menuHeight + this._gap < primary.y + primary.height)
-                    this._boxPointer._arrowSide = this._arrowSide = St.Side.TOP;
-                else if (primary.y + menuHeight + this._gap < sourceY)
-                    this._boxPointer._arrowSide = this._arrowSide = St.Side.BOTTOM;
-                else
-                    this._boxPointer._arrowSide = this._arrowSide = St.Side.TOP;
-            } else {
-                // horizontal submenu
-                if (sourceX + sourceWidth + menuWidth + this._gap < primary.x + primary.width)
-                    this._boxPointer._arrowSide = this._arrowSide = St.Side.LEFT;
-                else if (primary.x + menuWidth + this._gap < sourceX)
-                    this._boxPointer._arrowSide = this._arrowSide = St.Side.RIGHT;
-                else
-                    this._boxPointer._arrowSide = this._arrowSide = St.Side.LEFT;
-            }
-        }
-
-        this._boxPointer.setPosition(this.sourceActor, this._gap, this._alignment);
-
-        // Now show it
-        this.actor.reactive = true;
-        this._boxPointer.animateAppear();
-        this.isOpen = true;
-        this.emit('open-state-changed', true);
-    },
-
-    close: function() {
-        if (!this.isOpen)
-            return;
-
-        if (this._activeMenuItem)
-            this._activeMenuItem.setActive(false);
-        this.actor.reactive = false;
-        this._boxPointer.animateDisappear();
-        this.isOpen = false;
-        this.emit('open-state-changed', false);
-    },
-
     toggle: function() {
         if (this.isOpen)
-            this.close();
+            this.close(true);
         else
-            this.open();
-    },
-
-    _onKeyPressEvent: function(actor, event) {
-        // Move focus back to parent menu if the user types Left.
-        // (This handler is only connected if the PopupMenu is a
-        // submenu.)
-        if (this.isOpen &&
-            this._activeMenuItem &&
-            event.get_key_symbol() == Clutter.KEY_Left) {
-            this._activeMenuItem.setActive(false);
-            return true;
-        }
-
-        return false;
-    },
-
-    // return true if the actor is inside the menu or
-    // any actor related to the active submenu
-    contains: function(actor) {
-        if (this.actor.contains(actor))
-            return true;
-        if (this._activeMenuItem)
-            return this._activeMenuItem.contains(actor);
-        return false;
+            this.open(true);
     },
 
     destroy: function() {
@@ -836,7 +747,184 @@ PopupMenu.prototype = {
         this.emit('destroy');
     }
 };
-Signals.addSignalMethods(PopupMenu.prototype);
+Signals.addSignalMethods(PopupMenuBase.prototype);
+
+function PopupMenu() {
+    this._init.apply(this, arguments);
+}
+
+PopupMenu.prototype = {
+    __proto__: PopupMenuBase.prototype,
+
+    _init: function(sourceActor, alignment, arrowSide, gap) {
+        PopupMenuBase.prototype._init.call (this, sourceActor, 'popup-menu-content');
+
+        this._alignment = alignment;
+        this._arrowSide = arrowSide;
+        this._gap = gap;
+
+        this._boxPointer = new BoxPointer.BoxPointer(arrowSide,
+                                                     { x_fill: true,
+                                                       y_fill: true,
+                                                       x_align: St.Align.START });
+        this.actor = this._boxPointer.actor;
+        this.actor._delegate = this;
+        this.actor.style_class = 'popup-menu-boxpointer';
+        this._boxWrapper = new Shell.GenericContainer();
+        this._boxWrapper.connect('get-preferred-width', Lang.bind(this, this._boxGetPreferredWidth));
+        this._boxWrapper.connect('get-preferred-height', Lang.bind(this, this._boxGetPreferredHeight));
+        this._boxWrapper.connect('allocate', Lang.bind(this, this._boxAllocate));
+        this._boxPointer.bin.set_child(this._boxWrapper);
+        this._boxWrapper.add_actor(this.box);
+        this.actor.add_style_class_name('popup-menu');
+
+        global.focus_manager.add_group(this.actor);
+        this.actor.reactive = true;
+    },
+
+    _boxGetPreferredWidth: function (actor, forHeight, alloc) {
+        let columnWidths = this.getColumnWidths();
+        this.setColumnWidths(columnWidths);
+
+        // Now they will request the right sizes
+        [alloc.min_size, alloc.natural_size] = this.box.get_preferred_width(forHeight);
+    },
+
+    _boxGetPreferredHeight: function (actor, forWidth, alloc) {
+        [alloc.min_size, alloc.natural_size] = this.box.get_preferred_height(forWidth);
+    },
+
+    _boxAllocate: function (actor, box, flags) {
+        this.box.allocate(box, flags);
+    },
+
+    setArrowOrigin: function(origin) {
+        this._boxPointer.setArrowOrigin(origin);
+    },
+
+    open: function(animate) {
+        if (this.isOpen)
+            return;
+
+        this.isOpen = true;
+
+        this._boxPointer.setPosition(this.sourceActor, this._gap, this._alignment);
+        this._boxPointer.show(animate);
+
+        this.emit('open-state-changed', true);
+    },
+
+    close: function(animate) {
+        if (!this.isOpen)
+            return;
+
+        if (this._activeMenuItem)
+            this._activeMenuItem.setActive(false);
+
+        this._boxPointer.hide(animate);
+
+        this.isOpen = false;
+        this.emit('open-state-changed', false);
+    }
+};
+
+function PopupSubMenu() {
+    this._init.apply(this, arguments);
+}
+
+PopupSubMenu.prototype = {
+    __proto__: PopupMenuBase.prototype,
+
+    _init: function(sourceActor, sourceArrow) {
+        PopupMenuBase.prototype._init.call(this, sourceActor, 'popup-sub-menu');
+
+        this._arrow = sourceArrow;
+        this._arrow.rotation_center_z_gravity = Clutter.Gravity.CENTER;
+
+        this.actor = this.box;
+        this.actor._delegate = this;
+        this.actor.clip_to_allocation = true;
+        this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
+        this.actor.hide();
+    },
+
+    open: function(animate) {
+        if (this.isOpen)
+            return;
+
+        this.isOpen = true;
+
+        // we don't implement the !animate case because that doesn't
+        // currently get used...
+
+        let [naturalHeight, minHeight] = this.actor.get_preferred_height(-1);
+        this.actor.height = 0;
+        this.actor.show();
+        this.actor._arrow_rotation = this._arrow.rotation_angle_z;
+        Tweener.addTween(this.actor,
+                         { _arrow_rotation: 90,
+                           height: naturalHeight,
+                           time: 0.25,
+                           onUpdateScope: this,
+                           onUpdate: function() {
+                               this._arrow.rotation_angle_z = this.actor._arrow_rotation;
+                           },
+                           onCompleteScope: this,
+                           onComplete: function() {
+                               this.actor.set_height(-1);
+                               this.emit('open-state-changed', true);
+                           }
+                         });
+    },
+
+    close: function(animate) {
+        if (!this.isOpen)
+            return;
+
+        this.isOpen = false;
+
+        if (this._activeMenuItem)
+            this._activeMenuItem.setActive(false);
+
+        if (animate) {
+            this.actor._arrow_rotation = this._arrow.rotation_angle_z;
+            Tweener.addTween(this.actor,
+                             { _arrow_rotation: 0,
+                               height: 0,
+                               time: 0.25,
+                               onCompleteScope: this,
+                               onComplete: function() {
+                                   this.actor.hide();
+                                   this.actor.set_height(-1);
+
+                                   this.emit('open-state-changed', false);
+                               },
+                               onUpdateScope: this,
+                               onUpdate: function() {
+                                   this._arrow.rotation_angle_z = this.actor._arrow_rotation;
+                               }
+                             });
+            } else {
+                this._arrow.rotation_angle_z = 0;
+                this.actor.hide();
+
+                this.isOpen = false;
+                this.emit('open-state-changed', false);
+            }
+    },
+
+    _onKeyPressEvent: function(actor, event) {
+        // Move focus back to parent menu if the user types Left.
+
+        if (this.isOpen && event.get_key_symbol() == Clutter.KEY_Left) {
+            this.close(true);
+            this.sourceActor._delegate.setActive(true);
+            return true;
+        }
+
+        return false;
+    }
+};
 
 function PopupSubMenuMenuItem() {
     this._init.apply(this, arguments);
@@ -846,77 +934,46 @@ PopupSubMenuMenuItem.prototype = {
     __proto__: PopupBaseMenuItem.prototype,
 
     _init: function(text) {
-        PopupBaseMenuItem.prototype._init.call(this, { activate: false, hover: false });
-        this.actor.connect('enter-event', Lang.bind(this, this._mouseEnter));
-        this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
+        PopupBaseMenuItem.prototype._init.call(this);
+
+        this.actor.add_style_class_name('popup-submenu-menu-item');
 
         this.label = new St.Label({ text: text });
         this.addActor(this.label);
-        this.addActor(new St.Label({ text: '>' }));
+        this._triangle = new St.Label({ text: '\u25B8' });
+        this.addActor(this._triangle, { align: St.Align.END });
 
-        this.menu = new PopupMenu(this.actor, St.Align.MIDDLE, St.Side.LEFT, 0, true);
-        Main.chrome.addActor(this.menu.actor, { visibleInOverview: true,
-                                                affectsStruts: false });
-        this.menu.actor.hide();
-
-        this._openStateChangedId = this.menu.connect('open-state-changed', Lang.bind(this, this._subMenuOpenStateChanged));
-        this._activateId = this.menu.connect('activate', Lang.bind(this, this._subMenuActivate));
+        this.menu = new PopupSubMenu(this.actor, this._triangle);
+        this.menu.connect('open-state-changed', Lang.bind(this, this._subMenuOpenStateChanged));
     },
 
     _subMenuOpenStateChanged: function(menu, open) {
-        PopupBaseMenuItem.prototype.setActive.call(this, open);
-    },
-
-    _subMenuActivate: function(menu, menuItem) {
-        this.emit('activate', null);
-    },
-
-    setMenu: function(newmenu) {
-        if (this.menu) {
-            this.menu.close();
-            this.menu.disconnect(this._openStateChangedId);
-            this.menu.disconnect(this._activateId);
-        }
-        if (newmenu) {
-            this._openStateChangedId = newmenu.connect('open-state-changed', Lang.bind(this, this._subMenuOpenStateChanged));
-            this._activateId = newmenu.connect('activate', Lang.bind(this, this._subMenuActivate));
-        }
-        this.menu = newmenu;
+        if (open)
+            this.actor.add_style_pseudo_class('open');
+        else
+            this.actor.remove_style_pseudo_class('open');
     },
 
     destroy: function() {
-        if (this.menu)
-            this.menu.destroy();
+        this.menu.destroy();
         PopupBaseMenuItem.prototype.destroy.call(this);
     },
 
-    setActive: function(active) {
-        if (this.menu) {
-            if (active)
-                this.menu.open();
-            else
-                this.menu.close();
-        }
-
-        PopupBaseMenuItem.prototype.setActive.call(this, active);
-    },
-
     _onKeyPressEvent: function(actor, event) {
-        if (!this.menu)
-            return false;
         if (event.get_key_symbol() == Clutter.KEY_Right) {
+            this.menu.open(true);
             this.menu.activateFirst();
             return true;
         }
-        return false;
+        return PopupBaseMenuItem.prototype._onKeyPressEvent.call(this, actor, event);
     },
 
-    contains: function(actor) {
-        return this.menu && this.menu.contains(actor);
+    activate: function(event) {
+        this.menu.open(true);
     },
 
-    _mouseEnter: function(event) {
-        this.setActive(true);
+    _onButtonReleaseEvent: function(actor) {
+        this.menu.toggle();
     }
 };
 
@@ -939,23 +996,24 @@ PopupMenuManager.prototype = {
         this._leaveEventId = 0;
         this._activeMenu = null;
         this._menus = [];
-        this._delayedMenus = [];
+        this._preGrabInputMode = null;
     },
 
     addMenu: function(menu, position) {
         let menudata = {
             menu:              menu,
             openStateChangeId: menu.connect('open-state-changed', Lang.bind(this, this._onMenuOpenState)),
-            activateId:        menu.connect('activate', Lang.bind(this, this._onMenuActivated)),
             destroyId:         menu.connect('destroy', Lang.bind(this, this._onMenuDestroy)),
             enterId:           0,
-            focusId:           0
+            focusInId:         0,
+            focusOutId:        0
         };
 
         let source = menu.sourceActor;
         if (source) {
             menudata.enterId = source.connect('enter-event', Lang.bind(this, function() { this._onMenuSourceEnter(menu); }));
-            menudata.focusId = source.connect('key-focus-in', Lang.bind(this, function() { this._onMenuSourceEnter(menu); }));
+            menudata.focusInId = source.connect('key-focus-in', Lang.bind(this, function() { this._onMenuSourceEnter(menu); }));
+            menudata.focusOutId = source.connect('key-focus-out', Lang.bind(this, function() { this._onKeyFocusOut(menu); }));
         }
 
         if (position == undefined)
@@ -974,18 +1032,19 @@ PopupMenuManager.prototype = {
 
         let menudata = this._menus[position];
         menu.disconnect(menudata.openStateChangeId);
-        menu.disconnect(menudata.activateId);
         menu.disconnect(menudata.destroyId);
 
         if (menudata.enterId)
             menu.sourceActor.disconnect(menudata.enterId);
-        if (menudata.focusId)
-            menu.sourceActor.disconnect(menudata.focusId);
+        if (menudata.focusInId)
+            menu.sourceActor.disconnect(menudata.focusInId);
+        if (menudata.focusOutId)
+            menu.sourceActor.disconnect(menudata.focusOutId);
 
         this._menus.splice(position, 1);
     },
 
-    grab: function() {
+    _grab: function() {
         Main.pushModal(this._owner.actor);
 
         this._eventCaptureId = global.stage.connect('captured-event', Lang.bind(this, this._onEventCapture));
@@ -997,7 +1056,7 @@ PopupMenuManager.prototype = {
         this.grabbed = true;
     },
 
-    ungrab: function() {
+    _ungrab: function() {
         global.stage.disconnect(this._eventCaptureId);
         this._eventCaptureId = 0;
         global.stage.disconnect(this._keyPressEventId);
@@ -1013,40 +1072,100 @@ PopupMenuManager.prototype = {
 
     _onMenuOpenState: function(menu, open) {
         if (open) {
+            if (!this.grabbed) {
+                this._preGrabInputMode = global.stage_input_mode;
+                this._grab();
+            }
             this._activeMenu = menu;
-            if (!this.grabbed)
-                this.grab();
+
+            // if the focus is not already associated with the menu,
+            // then focus the menu
+            let focus = global.stage.key_focus;
+            if (!this._activeMenuContains(focus))
+                menu.sourceActor.grab_key_focus();
         } else if (menu == this._activeMenu) {
-            this._activeMenu = null;
+            let focus = global.stage.key_focus;
+            let fromActive = this._activeMenuContains(focus);
+
             if (this.grabbed)
-                this.ungrab();
+                this._ungrab();
+            this._activeMenu = null;
+
+            // If keynav was in effect before we grabbed, then we need
+            // to properly re-establish it after we ungrab. (popModal
+            // will have unset the focus.) If some part of the menu
+            // was focused at the time of the ungrab then focus its
+            // sourceActor. Otherwise just reset the focus to where it
+            // was right before the ungrab.
+            if (this._preGrabInputMode == Shell.StageInputMode.FOCUSED) {
+                global.stage_input_mode = Shell.StageInputMode.FOCUSED;
+                if (fromActive)
+                    menu.sourceActor.grab_key_focus();
+                else
+                    focus.grab_key_focus();
+            }
         }
     },
 
+    // change the currently-open menu without dropping grab
+    _changeMenu: function(newMenu) {
+        if (this._activeMenu) {
+            // _onOpenMenuState will drop the grab if it sees
+            // this._activeMenu being closed; so clear _activeMenu
+            // before closing it to keep that from happening
+            let oldMenu = this._activeMenu;
+            this._activeMenu = null;
+            oldMenu.close(false);
+            newMenu.open(false);
+        } else
+            newMenu.open(true);
+    },
+
     _onMenuSourceEnter: function(menu) {
         if (!this.grabbed || menu == this._activeMenu)
             return false;
 
-        if (this._activeMenu != null)
-            this._activeMenu.close();
-        menu.open();
+        this._changeMenu(menu);
         return false;
     },
 
-    _onMenuActivated: function(menu, item) {
-        if (this.grabbed)
-            this.ungrab();
+    _onKeyFocusOut: function(menu) {
+        if (!this.grabbed || menu != this._activeMenu)
+            return;
+
+        // We want to close the menu if the focus has moved somewhere
+        // other than inside the menu or to another menu's sourceActor.
+        // Unfortunately, when key-focus-out is emitted,
+        // stage.key_focus will be null. So we have to wait until
+        // after it emits the key-focus-in as well.
+        let id = global.stage.connect('notify::key-focus', Lang.bind(this,
+            function () {
+                global.stage.disconnect(id);
+
+                if (menu != this._activeMenu)
+                    return;
+
+                let focus = global.stage.key_focus;
+                if (!focus || this._activeMenuContains(focus))
+                    return;
+                if (focus._delegate && this._findMenu(focus._delegate.menu) != -1)
+                    return;
+                menu.close(true);
+            }));
     },
 
     _onMenuDestroy: function(menu) {
         this.removeMenu(menu);
     },
 
-    _eventIsOnActiveMenu: function(event) {
-        let src = event.get_source();
+    _activeMenuContains: function(actor) {
         return this._activeMenu != null
-                && (this._activeMenu.contains(src) ||
-                    (this._activeMenu.sourceActor && this._activeMenu.sourceActor.contains(src)));
+                && (this._activeMenu.actor.contains(actor) ||
+                    (this._activeMenu.sourceActor && this._activeMenu.sourceActor.contains(actor)));
+    },
+
+    _eventIsOnActiveMenu: function(event) {
+        return this._activeMenuContains(event.get_source());
     },
 
     _eventIsOnAnyMenuSource: function(event) {
@@ -1108,10 +1227,7 @@ PopupMenuManager.prototype = {
             let pos = this._findMenu(this._activeMenu);
             let next = this._menus[mod(pos + direction, this._menus.length)].menu;
             if (next != this._activeMenu) {
-                let oldMenu = this._activeMenu;
-                this._activeMenu = next;
-                oldMenu.close();
-                next.open();
+                this._changeMenu(next);
                 next.activateFirst();
             }
             return true;
@@ -1122,6 +1238,6 @@ PopupMenuManager.prototype = {
 
     _closeMenu: function() {
         if (this._activeMenu != null)
-            this._activeMenu.close();
+            this._activeMenu.close(true);
     }
 };
